@@ -61,13 +61,6 @@ static u64 highera(u64 x)  { return ((x + 0x8000) >> 32) & 0xffff; }
 static u64 highest(u64 x)  { return x >> 48; }
 static u64 highesta(u64 x) { return (x + 0x8000) >> 48; }
 
-static constexpr ScanAction toc_table[3][4] = {
-  // Absolute  Local    Imported data  Imported code
-  {  NONE,     BASEREL, DYNREL,        DYNREL },  // Shared object
-  {  NONE,     BASEREL, DYNREL,        DYNREL },  // Position-independent exec
-  {  NONE,     NONE,    DYNREL,        DYNREL },  // Position-dependent exec
-};
-
 // .plt is used only for lazy symbol resolution on PPC64. All PLT
 // calls are made via range extension thunks even if they are within
 // reach. Thunks read addresses from .got.plt and jump there.
@@ -169,12 +162,11 @@ void InputSection<E>::apply_reloc_alloc(Context<E> &ctx, u8 *base) {
         if (ctx.arg.apply_dynamic_relocs)
           *(ub64 *)loc = S + A;
       } else {
-        apply_dyn_absrel(ctx, sym, rel, loc, S, A, P, dynrel, toc_table);
+        apply_toc_rel(ctx, sym, rel, loc, S, A, P, dynrel);
       }
       break;
     case R_PPC64_TOC:
-      apply_dyn_absrel(ctx, *ctx.TOC, rel, loc, ctx.TOC->value, A, P, dynrel,
-                       toc_table);
+      apply_toc_rel(ctx, *ctx.TOC, rel, loc, ctx.TOC->value, A, P, dynrel);
       break;
     case R_PPC64_TOC16_HA:
       *(ub16 *)loc = ha(S + A - ctx.TOC->value);
@@ -250,13 +242,13 @@ void InputSection<E>::apply_reloc_alloc(Context<E> &ctx, u8 *base) {
       *(ub16 *)loc = ctx.got->get_tlsld_addr(ctx) - ctx.TOC->value;
       break;
     case R_PPC64_DTPREL16_HA:
-      *(ub16 *)loc = ha(S + A - ctx.tls_begin - E::tls_dtp_offset);
+      *(ub16 *)loc = ha(S + A - ctx.dtp_addr);
       break;
     case R_PPC64_TPREL16_HA:
       *(ub16 *)loc = ha(S + A - ctx.tp_addr);
       break;
     case R_PPC64_DTPREL16_LO:
-      *(ub16 *)loc = S + A - ctx.tls_begin - E::tls_dtp_offset;
+      *(ub16 *)loc = S + A - ctx.dtp_addr;
       break;
     case R_PPC64_TPREL16_LO:
       *(ub16 *)loc = S + A - ctx.tp_addr;
@@ -327,7 +319,7 @@ void InputSection<E>::apply_reloc_nonalloc(Context<E> &ctx, u8 *base) {
       break;
     }
     case R_PPC64_DTPREL64:
-      *(ub64 *)loc = S + A - ctx.tls_begin - E::tls_dtp_offset;
+      *(ub64 *)loc = S + A - ctx.dtp_addr;
       break;
     default:
       Fatal(ctx) << *this << ": apply_reloc_nonalloc: " << rel;
@@ -359,36 +351,37 @@ void InputSection<E>::scan_relocations(Context<E> &ctx) {
     }
 
     if (sym.is_ifunc())
-      sym.flags |= (NEEDS_GOT | NEEDS_PLT | NEEDS_OPD);
+      sym.flags.fetch_or(NEEDS_GOT | NEEDS_PLT | NEEDS_OPD,
+                         std::memory_order_relaxed);
 
     if (rel.r_type != R_PPC64_REL24 && sym.get_type() == STT_FUNC)
-      sym.flags |= NEEDS_OPD;
+      sym.flags.fetch_or(NEEDS_OPD, std::memory_order_relaxed);
 
     switch (rel.r_type) {
     case R_PPC64_ADDR64:
       if (sym.is_ifunc())
         this->file.num_dynrel++;
       else
-        scan_rel(ctx, sym, rel, toc_table);
+        scan_toc_rel(ctx, sym, rel);
       break;
     case R_PPC64_TOC:
-      scan_rel(ctx, sym, rel, toc_table);
+      scan_toc_rel(ctx, sym, rel);
       break;
     case R_PPC64_GOT_TPREL16_HA:
-      sym.flags |= NEEDS_GOTTP;
+      sym.flags.fetch_or(NEEDS_GOTTP, std::memory_order_relaxed);
       break;
     case R_PPC64_REL24:
       if (sym.is_imported)
-        sym.flags |= NEEDS_PLT;
+        sym.flags.fetch_or(NEEDS_PLT, std::memory_order_relaxed);
       break;
     case R_PPC64_PLT16_HA:
-      sym.flags |= NEEDS_GOT;
+      sym.flags.fetch_or(NEEDS_GOT, std::memory_order_relaxed);
       break;
     case R_PPC64_GOT_TLSGD16_HA:
-      sym.flags |= NEEDS_TLSGD;
+      sym.flags.fetch_or(NEEDS_TLSGD, std::memory_order_relaxed);
       break;
     case R_PPC64_GOT_TLSLD16_HA:
-      ctx.needs_tlsld = true;
+      ctx.needs_tlsld.store(true, std::memory_order_relaxed);
       break;
     case R_PPC64_REL64:
     case R_PPC64_TOC16_HA:
