@@ -23,6 +23,7 @@
 // TLS block as TP (with some addend). As a result, offsets from TP to
 // thread-local variables (TLVs) in the main executable are all negative.
 //
+// https://github.com/rui314/mold/wiki/i386-psabi.pdf
 // https://github.com/rui314/mold/wiki/x86-64-psabi.pdf
 
 #include "mold.h"
@@ -326,6 +327,7 @@ void InputSection<E>::apply_reloc_alloc(Context<E> &ctx, u8 *base) {
 
         switch (rels[i + 1].r_type) {
         case R_X86_64_PLT32:
+        case R_X86_64_PC32:
         case R_X86_64_GOTPCREL:
         case R_X86_64_GOTPCRELX: {
           // The original instructions are the following:
@@ -378,7 +380,8 @@ void InputSection<E>::apply_reloc_alloc(Context<E> &ctx, u8 *base) {
         // block, and we know the exact size of the TLS block we have
         // created, so we can just subtract it from TP.
         switch (rels[i + 1].r_type) {
-        case R_X86_64_PLT32: {
+        case R_X86_64_PLT32:
+        case R_X86_64_PC32: {
           // The original instructions are the following:
           //
           //  48 8d 3d 00 00 00 00    lea    foo@tlsld(%rip), %rdi
@@ -627,23 +630,23 @@ void InputSection<E>::scan_relocations(Context<E> &ctx) {
     }
 
     if (sym.is_ifunc())
-      sym.flags |= (NEEDS_GOT | NEEDS_PLT);
+      sym.flags.fetch_or(NEEDS_GOT | NEEDS_PLT, std::memory_order_relaxed);
 
     switch (rel.r_type) {
     case R_X86_64_8:
     case R_X86_64_16:
     case R_X86_64_32:
     case R_X86_64_32S:
-      scan_rel(ctx, sym, rel, absrel_table);
+      scan_absrel(ctx, sym, rel);
       break;
     case R_X86_64_64:
-      scan_rel(ctx, sym, rel, dyn_absrel_table);
+      scan_dyn_absrel(ctx, sym, rel);
       break;
     case R_X86_64_PC8:
     case R_X86_64_PC16:
     case R_X86_64_PC32:
     case R_X86_64_PC64:
-      scan_rel(ctx, sym, rel, pcrel_table);
+      scan_pcrel(ctx, sym, rel);
       break;
     case R_X86_64_GOT32:
     case R_X86_64_GOT64:
@@ -651,7 +654,7 @@ void InputSection<E>::scan_relocations(Context<E> &ctx) {
     case R_X86_64_GOTPC64:
     case R_X86_64_GOTPCREL:
     case R_X86_64_GOTPCREL64:
-      sym.flags |= NEEDS_GOT;
+      sym.flags.fetch_or(NEEDS_GOT, std::memory_order_relaxed);
       break;
     case R_X86_64_GOTPCRELX: {
       if (rel.r_addend != -4)
@@ -660,7 +663,7 @@ void InputSection<E>::scan_relocations(Context<E> &ctx) {
       bool do_relax = ctx.arg.relax && !sym.is_imported &&
                       sym.is_relative() && relax_gotpcrelx(loc - 2);
       if (!do_relax)
-        sym.flags |= NEEDS_GOT;
+        sym.flags.fetch_or(NEEDS_GOT, std::memory_order_relaxed);
       break;
     }
     case R_X86_64_REX_GOTPCRELX: {
@@ -670,13 +673,13 @@ void InputSection<E>::scan_relocations(Context<E> &ctx) {
       bool do_relax = ctx.arg.relax && !sym.is_imported &&
                       sym.is_relative() && relax_rex_gotpcrelx(loc - 3);
       if (!do_relax)
-        sym.flags |= NEEDS_GOT;
+        sym.flags.fetch_or(NEEDS_GOT, std::memory_order_relaxed);
       break;
     }
     case R_X86_64_PLT32:
     case R_X86_64_PLTOFF64:
       if (sym.is_imported)
-        sym.flags |= NEEDS_PLT;
+        sym.flags.fetch_or(NEEDS_PLT, std::memory_order_relaxed);
       break;
     case R_X86_64_TLSGD: {
       if (rel.r_addend != -4)
@@ -686,14 +689,15 @@ void InputSection<E>::scan_relocations(Context<E> &ctx) {
         Fatal(ctx) << *this << ": TLSGD reloc must be followed by PLT or GOTPCREL";
 
       if (u32 ty = rels[i + 1].r_type;
-          ty != R_X86_64_PLT32 && ty != R_X86_64_PLTOFF64 &&
-          ty != R_X86_64_GOTPCREL && ty != R_X86_64_GOTPCRELX)
+          ty != R_X86_64_PLT32 && ty != R_X86_64_PC32 &&
+          ty != R_X86_64_PLTOFF64 && ty != R_X86_64_GOTPCREL &&
+          ty != R_X86_64_GOTPCRELX)
         Fatal(ctx) << *this << ": TLSGD reloc must be followed by PLT or GOTPCREL";
 
       if (relax_tlsgd(ctx, sym))
         i++;
       else
-        sym.flags |= NEEDS_TLSGD;
+        sym.flags.fetch_or(NEEDS_TLSGD, std::memory_order_relaxed);
       break;
     }
     case R_X86_64_TLSLD: {
@@ -704,26 +708,27 @@ void InputSection<E>::scan_relocations(Context<E> &ctx) {
         Fatal(ctx) << *this << ": TLSLD reloc must be followed by PLT or GOTPCREL";
 
       if (u32 ty = rels[i + 1].r_type;
-          ty != R_X86_64_PLT32 && ty != R_X86_64_PLTOFF64 &&
-          ty != R_X86_64_GOTPCREL && ty != R_X86_64_GOTPCRELX)
+          ty != R_X86_64_PLT32 && ty != R_X86_64_PC32 &&
+          ty != R_X86_64_PLTOFF64 && ty != R_X86_64_GOTPCREL &&
+          ty != R_X86_64_GOTPCRELX)
         Fatal(ctx) << *this << ": TLSLD reloc must be followed by PLT or GOTPCREL";
 
       if (relax_tlsld(ctx))
         i++;
       else
-        ctx.needs_tlsld = true;
+        ctx.needs_tlsld.store(true, std::memory_order_relaxed);
       break;
     }
     case R_X86_64_GOTTPOFF: {
       if (rel.r_addend != -4)
         Fatal(ctx) << *this << ": bad r_addend for R_X86_64_GOTTPOFF";
 
-      ctx.has_gottp_rel = true;
+      ctx.has_gottp_rel.store(true, std::memory_order_relaxed);
 
       bool do_relax = ctx.arg.relax && !ctx.arg.shared &&
                       !sym.is_imported && relax_gottpoff(loc - 3);
       if (!do_relax)
-        sym.flags |= NEEDS_GOTTP;
+        sym.flags.fetch_or(NEEDS_GOTTP, std::memory_order_relaxed);
       break;
     }
     case R_X86_64_GOTPC32_TLSDESC: {
@@ -735,7 +740,7 @@ void InputSection<E>::scan_relocations(Context<E> &ctx) {
                    << " against an invalid code sequence";
 
       if (!relax_tlsdesc(ctx, sym))
-        sym.flags |= NEEDS_TLSDESC;
+        sym.flags.fetch_or(NEEDS_TLSDESC, std::memory_order_relaxed);
       break;
     }
     case R_X86_64_GOTOFF64:
