@@ -2,6 +2,7 @@
 
 #include <bit>
 #include <cstring>
+#include <tbb/parallel_sort.h>
 
 #ifndef _WIN32
 # include <unistd.h>
@@ -49,13 +50,21 @@ ElfShdr<E> *InputFile<E>::find_section(i64 type) {
   return nullptr;
 }
 
+// This function is actually racy, in that "tearing" might be observed
+// during the read or write of `sym->file`. However, on most platforms
+// read/write of pointers never tears, and even if it did, it would be
+// harmless, as `sym->file` will not become equal to `sym`, or even if
+// it did by chance, we just wipe a symbol twice.
+//
+// Doing this in a standard-compliant way is too cumbersome, hence
+// here we just pretend a compiler will never optimize this into a
+// form that actually exploits the UB.
 template <typename E>
+__attribute__((no_sanitize("thread")))
 void InputFile<E>::clear_symbols() {
-  for (Symbol<E> *sym : get_global_syms()) {
-    std::scoped_lock lock(sym->mu);
+  for (Symbol<E> *sym : get_global_syms())
     if (sym->file == this)
       sym->clear();
-  }
 }
 
 // Find the source filename. It should be listed in symtab as STT_FILE.
@@ -89,10 +98,9 @@ static bool is_debug_section(const ElfShdr<E> &shdr, std::string_view name) {
 }
 
 template <typename E>
-u32 ObjectFile<E>::read_note_gnu_property(Context<E> &ctx,
-                                          const ElfShdr<E> &shdr) {
+void
+ObjectFile<E>::read_note_gnu_property(Context<E> &ctx, const ElfShdr<E> &shdr) {
   std::string_view data = this->get_string(ctx, shdr);
-  u32 ret = 0;
 
   while (!data.empty()) {
     ElfNhdr<E> &hdr = *(ElfNhdr<E> *)data.data();
@@ -111,12 +119,20 @@ u32 ObjectFile<E>::read_note_gnu_property(Context<E> &ctx,
       u32 type = *(U32<E> *)desc.data();
       u32 size = *(U32<E> *)(desc.data() + 4);
       desc = desc.substr(8);
-      if (type == GNU_PROPERTY_X86_FEATURE_1_AND)
-        ret |= *(U32<E> *)desc.data();
+
+      // The majority of currently defined .note.gnu.property
+      // use 32-bit values.
+      // We don't know how to handle anything else, so if we encounter
+      // one, skip it.
+      //
+      // The following properties have a different size:
+      // - GNU_PROPERTY_STACK_SIZE
+      // - GNU_PROPERTY_NO_COPY_ON_PROTECTED
+      if (size == 4)
+        gnu_properties[type] |= *(U32<E> *)desc.data();
       desc = desc.substr(align_to(size, sizeof(Word<E>)));
     }
   }
-  return ret;
 }
 
 template <typename E>
@@ -126,7 +142,7 @@ void ObjectFile<E>::initialize_sections(Context<E> &ctx) {
     const ElfShdr<E> &shdr = this->elf_sections[i];
 
     if ((shdr.sh_flags & SHF_EXCLUDE) && !(shdr.sh_flags & SHF_ALLOC) &&
-        shdr.sh_type != SHT_LLVM_ADDRSIG)
+        shdr.sh_type != SHT_LLVM_ADDRSIG && !ctx.arg.relocatable)
       continue;
 
     switch (shdr.sh_type) {
@@ -134,8 +150,15 @@ void ObjectFile<E>::initialize_sections(Context<E> &ctx) {
       // Get the signature of this section group.
       if (shdr.sh_info >= this->elf_syms.size())
         Fatal(ctx) << *this << ": invalid symbol index";
-      const ElfSym<E> &sym = this->elf_syms[shdr.sh_info];
-      std::string_view signature = this->symbol_strtab.data() + sym.st_name;
+      const ElfSym<E> &esym = this->elf_syms[shdr.sh_info];
+
+      std::string_view signature;
+      if (esym.st_type == STT_SECTION) {
+        signature = this->shstrtab.data() +
+                    this->elf_sections[esym.st_shndx].sh_name;
+      } else {
+        signature = this->symbol_strtab.data() + esym.st_name;
+      }
 
       // Ignore a broken comdat group GCC emits for .debug_macros.
       // https://github.com/rui314/mold/issues/438
@@ -155,7 +178,7 @@ void ObjectFile<E>::initialize_sections(Context<E> &ctx) {
       typename decltype(ctx.comdat_groups)::const_accessor acc;
       ctx.comdat_groups.insert(acc, {signature, ComdatGroup()});
       ComdatGroup *group = const_cast<ComdatGroup *>(&acc->second);
-      comdat_groups.push_back({group, entries.subspan(1)});
+      comdat_groups.push_back({group, (u32)i, entries.subspan(1)});
       break;
     }
     case SHT_SYMTAB_SHNDX:
@@ -175,7 +198,7 @@ void ObjectFile<E>::initialize_sections(Context<E> &ctx) {
       // area in GNU linkers. We ignore that section because silently
       // making the stack area executable is too dangerous. Tell our
       // users about the difference if that matters.
-      if (name == ".note.GNU-stack") {
+      if (name == ".note.GNU-stack" && !ctx.arg.relocatable) {
         if (shdr.sh_flags & SHF_EXECINSTR) {
           if (!ctx.arg.z_execstack && !ctx.arg.z_execstack_if_needed)
             Warn(ctx) << *this << ": this file may cause a segmentation"
@@ -188,7 +211,7 @@ void ObjectFile<E>::initialize_sections(Context<E> &ctx) {
       }
 
       if (name == ".note.gnu.property") {
-        this->features = read_note_gnu_property(ctx, shdr);
+        read_note_gnu_property(ctx, shdr);
         continue;
       }
 
@@ -213,11 +236,18 @@ void ObjectFile<E>::initialize_sections(Context<E> &ctx) {
         continue;
 
       // Save .llvm_addrsig for --icf=safe.
-      if (shdr.sh_type == SHT_LLVM_ADDRSIG) {
+      if (shdr.sh_type == SHT_LLVM_ADDRSIG && !ctx.arg.relocatable) {
         llvm_addrsig = std::make_unique<InputSection<E>>(ctx, *this, name, i);
         continue;
       }
 
+      // If an output file doesn't have a section header (i.e.
+      // --oformat=binary is given), we discard all non-memory-allocated
+      // sections. This is because without a section header, we can't find
+      // their places in an output file in the first place.
+      if (ctx.arg.oformat_binary && !(shdr.sh_flags & SHF_ALLOC))
+        continue;
+
       this->sections[i] = std::make_unique<InputSection<E>>(ctx, *this, name, i);
 
       // Save debug sections for --gdb-index.
@@ -439,7 +469,7 @@ void ObjectFile<E>::initialize_symbols(Context<E> &ctx) {
 
     std::string_view name;
     if (esym.st_type == STT_SECTION)
-      name = this->shstrtab.data() + this->elf_sections[esym.st_shndx].sh_name;
+      name = this->shstrtab.data() + this->elf_sections[get_shndx(esym)].sh_name;
     else
       name = this->symbol_strtab.data() + esym.st_name;
 
@@ -515,7 +545,7 @@ static size_t find_null(std::string_view data, u64 entsize) {
     return data.find('\0');
 
   for (i64 i = 0; i <= data.size() - entsize; i += entsize)
-    if (data.substr(i, i + entsize).find_first_not_of('\0') == data.npos)
+    if (data.substr(i, entsize).find_first_not_of('\0') == data.npos)
       return i;
 
   return data.npos;
@@ -541,9 +571,16 @@ static size_t find_null(std::string_view data, u64 entsize) {
 template <typename E>
 static std::unique_ptr<MergeableSection<E>>
 split_section(Context<E> &ctx, InputSection<E> &sec) {
+  if (!sec.is_alive || sec.sh_size == 0 || sec.relsec_idx != -1)
+    return nullptr;
+
+  const ElfShdr<E> &shdr = sec.shdr();
+  if (!(shdr.sh_flags & SHF_MERGE))
+    return nullptr;
+
   std::unique_ptr<MergeableSection<E>> rec(new MergeableSection<E>);
-  rec->parent = MergedSection<E>::get_instance(ctx, sec.name(), sec.shdr().sh_type,
-                                               sec.shdr().sh_flags);
+  rec->parent = MergedSection<E>::get_instance(ctx, sec.name(), shdr.sh_type,
+                                               shdr.sh_flags);
   rec->p2align = sec.p2align;
 
   // If thes section contents are compressed, uncompress them.
@@ -551,11 +588,18 @@ split_section(Context<E> &ctx, InputSection<E> &sec) {
 
   std::string_view data = sec.contents;
   const char *begin = data.data();
-  u64 entsize = sec.shdr().sh_entsize;
+  u64 entsize = shdr.sh_entsize;
   HyperLogLog estimator;
 
   // Split sections
-  if (sec.shdr().sh_flags & SHF_STRINGS) {
+  if (shdr.sh_flags & SHF_STRINGS) {
+    if (entsize == 0) {
+      // GHC (Glasgow Haskell Compiler) sometimes creates a mergeable
+      // string section with entsize of 0 instead of 1, though such
+      // entsize is technically wrong. This is a workaround for the issue.
+      entsize = 1;
+    }
+
     while (!data.empty()) {
       size_t end = find_null(data, entsize);
       if (end == data.npos)
@@ -572,6 +616,11 @@ split_section(Context<E> &ctx, InputSection<E> &sec) {
       estimator.insert(hash);
     }
   } else {
+    // OCaml compiler seems to create a mergeable non-string section with
+    // entisze of 0. Such section is malformed. We do not split such section.
+    if (entsize == 0)
+      return nullptr;
+
     if (data.size() % entsize)
       Fatal(ctx) << sec << ": section size is not multiple of sh_entsize";
 
@@ -640,18 +689,17 @@ void ObjectFile<E>::initialize_mergeable_sections(Context<E> &ctx) {
   mergeable_sections.resize(sections.size());
 
   for (i64 i = 0; i < sections.size(); i++) {
-    std::unique_ptr<InputSection<E>> &isec = sections[i];
-    if (isec && isec->is_alive && (isec->shdr().sh_flags & SHF_MERGE) &&
-        isec->sh_size && isec->shdr().sh_entsize &&
-        isec->relsec_idx == -1) {
-      mergeable_sections[i] = split_section(ctx, *isec);
-      isec->is_alive = false;
+    if (std::unique_ptr<InputSection<E>> &isec = sections[i]) {
+      if (std::unique_ptr<MergeableSection<E>> m = split_section(ctx, *isec)) {
+        mergeable_sections[i] = std::move(m);
+        isec->is_alive = false;
+      }
     }
   }
 }
 
 template <typename E>
-void ObjectFile<E>::register_section_pieces(Context<E> &ctx) {
+void ObjectFile<E>::resolve_section_pieces(Context<E> &ctx) {
   for (std::unique_ptr<MergeableSection<E>> &m : mergeable_sections) {
     if (m) {
       m->fragments.reserve(m->strings.size());
@@ -716,11 +764,11 @@ void ObjectFile<E>::register_section_pieces(Context<E> &ctx) {
       if (!m)
         continue;
 
-      i64 r_addend = isec->get_addend(r);
+      i64 r_addend = get_addend(*isec, r);
 
       SectionFragment<E> *frag;
-      i64 frag_offset;
-      std::tie(frag, frag_offset) = m->get_fragment(esym.st_value + r_addend);
+      i64 in_frag_offset;
+      std::tie(frag, in_frag_offset) = m->get_fragment(esym.st_value + r_addend);
 
       if (!frag)
         Fatal(ctx) << *this << ": bad relocation at " << r.r_sym;
@@ -731,8 +779,7 @@ void ObjectFile<E>::register_section_pieces(Context<E> &ctx) {
       sym.sym_idx = r.r_sym;
       sym.visibility = STV_HIDDEN;
       sym.set_frag(frag);
-      sym.value = frag_offset - r_addend;
-
+      sym.value = in_frag_offset - r_addend;
       r.r_sym = this->elf_syms.size() + idx;
       idx++;
     }
@@ -787,7 +834,6 @@ void ObjectFile<E>::parse(Context<E> &ctx) {
   initialize_sections(ctx);
   initialize_symbols(ctx);
   sort_relocations(ctx);
-  initialize_mergeable_sections(ctx);
   initialize_ehframe_sections(ctx);
 }
 
@@ -812,22 +858,12 @@ static u64 get_rank(InputFile<E> *file, const ElfSym<E> &esym, bool is_lazy) {
     return (5 << 24) + file->priority;
   }
 
-  // GCC creates symbols in COMDATs with STB_GNU_UNIQUE instead of
-  // STB_WEAK if it was configured to do so at build time or the
-  // -fgnu-unique flag was given. In order to to not select a
-  // GNU_UNIQUE symbol in a discarded COMDAT section, we treat it as
-  // if it were weak.
-  //
-  // It looks like STB_GNU_UNIQUE is not a popular option anymore and
-  // often disabled by default though.
-  bool is_weak = (esym.st_bind == STB_WEAK || esym.st_bind == STB_GNU_UNIQUE);
-
   if (file->is_dso || is_lazy) {
-    if (is_weak)
+    if (esym.st_bind == STB_WEAK)
       return (4 << 24) + file->priority;
     return (3 << 24) + file->priority;
   }
-  if (is_weak)
+  if (esym.st_bind == STB_WEAK)
     return (2 << 24) + file->priority;
   return (1 << 24) + file->priority;
 }
@@ -891,7 +927,7 @@ void ObjectFile<E>::resolve_symbols(Context<E> &ctx) {
     InputSection<E> *isec = nullptr;
     if (!esym.is_abs() && !esym.is_common()) {
       isec = get_section(esym);
-      if (!isec)
+      if (!isec || !isec->is_alive)
         continue;
     }
 
@@ -930,12 +966,11 @@ ObjectFile<E>::mark_live_objects(Context<E> &ctx,
     if (esym.is_weak())
       continue;
 
-    std::scoped_lock lock(sym.mu);
     if (!sym.file)
       continue;
 
     bool keep = esym.is_undef() || (esym.is_common() && !sym.esym().is_common());
-    if (keep && !sym.file->is_alive.exchange(true)) {
+    if (keep && fast_mark(sym.file->is_alive)) {
       feeder(sym.file);
 
       if (sym.traced)
@@ -957,24 +992,17 @@ ObjectFile<E>::mark_live_objects(Context<E> &ctx,
 // contains only a single copy of `std::vector<int>`.
 template <typename E>
 void ObjectFile<E>::resolve_comdat_groups() {
-  for (auto &pair : comdat_groups) {
-    ComdatGroup *group = pair.first;
-    update_minimum(group->owner, this->priority);
-  }
+  for (ComdatGroupRef<E> &ref : comdat_groups)
+    update_minimum(ref.group->owner, this->priority);
 }
 
 template <typename E>
 void ObjectFile<E>::eliminate_duplicate_comdat_groups() {
-  for (auto &pair : comdat_groups) {
-    ComdatGroup *group = pair.first;
-    if (group->owner == this->priority)
-      continue;
-
-    std::span<U32<E>> entries = pair.second;
-    for (u32 i : entries)
-      if (sections[i])
-        sections[i]->kill();
-  }
+  for (ComdatGroupRef<E> &ref : comdat_groups)
+    if (ref.group->owner != this->priority)
+      for (u32 i : ref.members)
+        if (sections[i])
+          sections[i]->kill();
 }
 
 template <typename E>
@@ -1026,51 +1054,58 @@ void ObjectFile<E>::claim_unresolved_symbols(Context<E> &ctx) {
       }
     }
 
-    auto claim = [&] {
+    auto claim = [&](bool is_imported) {
+      if (sym.traced)
+        SyncOut(ctx) << "trace-symbol: " << *this << ": unresolved"
+                     << (esym.is_weak() ? " weak" : "")
+                     << " symbol " << sym;
+
       sym.file = this;
       sym.origin = 0;
       sym.value = 0;
       sym.sym_idx = i;
       sym.is_weak = false;
+      sym.is_imported = is_imported;
       sym.is_exported = false;
+      sym.ver_idx = is_imported ? 0 : ctx.default_version;
     };
 
+    if (esym.is_undef_weak()) {
+      if (ctx.arg.shared && sym.visibility != STV_HIDDEN &&
+          ctx.arg.z_dynamic_undefined_weak) {
+        // Global weak undefined symbols are promoted to dynamic symbols
+        // when when linking a DSO, unless `-z nodynamic_undefined_weak`
+        // was given.
+        claim(true);
+      } else {
+        // Otherwise, weak undefs are converted to absolute symbols with value 0.
+        claim(false);
+      }
+      continue;
+    }
+
     if (ctx.arg.unresolved_symbols == UNRESOLVED_WARN)
       report_undef(sym);
 
-    // Convert remaining undefined symbols to dynamic symbols.
-    if (ctx.arg.shared && sym.visibility != STV_HIDDEN) {
-      // Traditionally, remaining undefined symbols cause a link failure
-      // only when we are creating an executable. Undefined symbols in
-      // shared objects are promoted to dynamic symbols, so that they'll
-      // get another chance to be resolved at run-time. You can change the
-      // behavior by passing `-z defs` to the linker.
-      //
-      // Even if `-z defs` is given, weak undefined symbols are still
-      // promoted to dynamic symbols for compatibility with other linkers.
-      // Some major programs, notably Firefox, depend on the behavior
-      // (they use this loophole to export symbols from libxul.so).
-      if (!ctx.arg.z_defs || esym.is_undef_weak() ||
-          ctx.arg.unresolved_symbols != UNRESOLVED_ERROR) {
-        claim();
-        sym.ver_idx = 0;
-        sym.is_imported = true;
-
-        if (sym.traced)
-          SyncOut(ctx) << "trace-symbol: " << *this << ": unresolved"
-                       << (esym.is_weak() ? " weak" : "")
-                       << " symbol " << sym;
-        continue;
-      }
+    // Traditionally, remaining undefined symbols cause a link failure
+    // only when we are creating an executable. Undefined symbols in
+    // shared objects are promoted to dynamic symbols, so that they'll
+    // get another chance to be resolved at run-time. You can change the
+    // behavior by passing `-z defs` to the linker.
+    //
+    // Even if `-z defs` is given, weak undefined symbols are still
+    // promoted to dynamic symbols for compatibility with other linkers.
+    // Some major programs, notably Firefox, depend on the behavior
+    // (they use this loophole to export symbols from libxul.so).
+    if (ctx.arg.shared && sym.visibility != STV_HIDDEN &&
+        (!ctx.arg.z_defs || ctx.arg.unresolved_symbols != UNRESOLVED_ERROR)) {
+      claim(true);
+      continue;
     }
 
     // Convert remaining undefined symbols to absolute symbols with value 0.
-    if (ctx.arg.unresolved_symbols != UNRESOLVED_ERROR ||
-        ctx.arg.noinhibit_exec || esym.is_undef_weak()) {
-      claim();
-      sym.ver_idx = ctx.default_version;
-      sym.is_imported = false;
-    }
+    if (ctx.arg.unresolved_symbols != UNRESOLVED_ERROR || ctx.arg.noinhibit_exec)
+      claim(false);
   }
 }
 
@@ -1120,14 +1155,6 @@ void ObjectFile<E>::convert_common_symbols(Context<E> &ctx) {
   if (!has_common_symbol)
     return;
 
-  OutputSection<E> *common =
-    OutputSection<E>::get_instance(ctx, ".common", SHT_NOBITS,
-                                   SHF_WRITE | SHF_ALLOC);
-
-  OutputSection<E> *tls_common =
-    OutputSection<E>::get_instance(ctx, ".tls_common", SHT_NOBITS,
-                                   SHF_WRITE | SHF_ALLOC | SHF_TLS);
-
   for (i64 i = this->first_global; i < this->elf_syms.size(); i++) {
     if (!this->elf_syms[i].is_common())
       continue;
@@ -1145,18 +1172,23 @@ void ObjectFile<E>::convert_common_symbols(Context<E> &ctx) {
     ElfShdr<E> &shdr = elf_sections2.back();
     memset(&shdr, 0, sizeof(shdr));
 
-    bool is_tls = (sym.get_type() == STT_TLS);
-    shdr.sh_flags = is_tls ? (SHF_ALLOC | SHF_TLS) : SHF_ALLOC;
+    std::string_view name;
+
+    if (sym.get_type() == STT_TLS) {
+      name = ".tls_common";
+      shdr.sh_flags = SHF_ALLOC | SHF_WRITE | SHF_TLS;
+    } else {
+      name = ".common";
+      shdr.sh_flags = SHF_ALLOC | SHF_WRITE;
+    }
+
     shdr.sh_type = SHT_NOBITS;
     shdr.sh_size = this->elf_syms[i].st_size;
     shdr.sh_addralign = this->elf_syms[i].st_value;
 
     i64 idx = this->elf_sections.size() + elf_sections2.size() - 1;
     std::unique_ptr<InputSection<E>> isec =
-      std::make_unique<InputSection<E>>(ctx, *this,
-                                        is_tls ? ".tls_common" : ".common",
-                                        idx);
-    isec->output_section = is_tls ? tls_common : common;
+      std::make_unique<InputSection<E>>(ctx, *this, name, idx);
 
     sym.file = this;
     sym.set_input_section(isec.get());
@@ -1199,6 +1231,8 @@ void ObjectFile<E>::compute_symtab_size(Context<E> &ctx) {
   if (ctx.arg.strip_all)
     return;
 
+  this->output_sym_indices.resize(this->elf_syms.size(), -1);
+
   auto is_alive = [&](Symbol<E> &sym) -> bool {
     if (!ctx.arg.gc_sections)
       return true;
@@ -1217,7 +1251,7 @@ void ObjectFile<E>::compute_symtab_size(Context<E> &ctx) {
 
       if (is_alive(sym) && should_write_to_local_symtab(ctx, sym)) {
         this->strtab_size += sym.name().size() + 1;
-        this->num_local_symtab++;
+        this->output_sym_indices[i] = this->num_local_symtab++;
         sym.write_to_symtab = true;
       }
     }
@@ -1232,10 +1266,10 @@ void ObjectFile<E>::compute_symtab_size(Context<E> &ctx) {
       this->strtab_size += sym.name().size() + 1;
       // Global symbols can be demoted to local symbols based on visibility,
       // version scripts etc.
-      if (sym.is_local())
-        this->num_local_symtab++;
+      if (sym.is_local(ctx))
+        this->output_sym_indices[i] = this->num_local_symtab++;
       else
-        this->num_global_symtab++;
+        this->output_sym_indices[i] = this->num_global_symtab++;
       sym.write_to_symtab = true;
     }
   }
@@ -1249,11 +1283,13 @@ void ObjectFile<E>::populate_symtab(Context<E> &ctx) {
   i64 strtab_off = this->strtab_offset;
 
   auto write_sym = [&](Symbol<E> &sym, i64 &symtab_idx) {
-    ElfSym<E> &esym = symtab_base[symtab_idx++];
-    esym = to_output_esym(ctx, sym);
-    esym.st_name = strtab_off;
-    write_string(strtab_base + strtab_off, sym.name());
-    strtab_off += sym.name().size() + 1;
+    U32<E> *xindex = nullptr;
+    if (ctx.symtab_shndx)
+      xindex = (U32<E> *)(ctx.buf + ctx.symtab_shndx->shdr.sh_offset +
+                          symtab_idx * 4);
+
+    symtab_base[symtab_idx++] = to_output_esym(ctx, sym, strtab_off, xindex);
+    strtab_off += write_string(strtab_base + strtab_off, sym.name());
   };
 
   i64 local_symtab_idx = this->local_symtab_idx;
@@ -1267,7 +1303,7 @@ void ObjectFile<E>::populate_symtab(Context<E> &ctx) {
   for (i64 i = this->first_global; i < this->elf_syms.size(); i++) {
     Symbol<E> &sym = *this->symbols[i];
     if (sym.file == this && sym.write_to_symtab) {
-      if (sym.is_local())
+      if (sym.is_local(ctx))
         write_sym(sym, local_symtab_idx);
       else
         write_sym(sym, global_symtab_idx);
@@ -1385,7 +1421,7 @@ void SharedFile<E>::parse(Context<E> &ctx) {
 // uses `posix_spawn` is linked either to that of "GLIBC_2.15" or that of
 // "GLIBC_2.2.5"
 //
-// Versions are just stirngs, and no ordering is defined between them.
+// Versions are just strings, and no ordering is defined between them.
 // For example, "GLIBC_2.15" is not considered a newer version of
 // "GLIBC_2.2.5" or vice versa. They are considered just different.
 //
@@ -1459,7 +1495,7 @@ SharedFile<E>::mark_live_objects(Context<E> &ctx,
       print_trace_symbol(ctx, *this, esym, sym);
 
     if (esym.is_undef() && sym.file && !sym.file->is_dso &&
-        !sym.file->is_alive.exchange(true)) {
+        fast_mark(sym.file->is_alive)) {
       feeder(sym.file);
 
       if (sym.traced)
@@ -1472,11 +1508,29 @@ SharedFile<E>::mark_live_objects(Context<E> &ctx,
 template <typename E>
 std::vector<Symbol<E> *> SharedFile<E>::find_aliases(Symbol<E> *sym) {
   assert(sym->file == this);
-  std::vector<Symbol<E> *> vec;
-  for (Symbol<E> *sym2 : this->symbols)
-    if (sym2->file == this && sym != sym2 &&
-        sym->esym().st_value == sym2->esym().st_value)
-      vec.push_back(sym2);
+
+  std::call_once(init_aliases, [&] {
+    for (Symbol<E> *sym : this->symbols)
+      if (sym->file == this)
+        aliases.push_back(sym);
+
+    tbb::parallel_sort(aliases.begin(), aliases.end(),
+                       [](Symbol<E> *a, Symbol<E> *b) {
+      const ElfSym<E> &x = a->esym();
+      const ElfSym<E> &y = b->esym();
+      return std::tuple{x.st_value, &x} < std::tuple{y.st_value, &y};
+    });
+  });
+
+  struct Cmp {
+    bool operator()(Symbol<E> *sym, u64 val) { return sym->esym().st_value < val; }
+    bool operator()(u64 val, Symbol<E> *sym) { return val < sym->esym().st_value; }
+  };
+
+  auto [begin, end] =
+    std::equal_range(aliases.begin(), aliases.end(), sym->esym().st_value, Cmp{});
+  std::vector<Symbol<E> *> vec{begin, end};
+  std::erase(vec, sym);
   return vec;
 }
 
@@ -1490,11 +1544,10 @@ std::vector<Symbol<E> *> SharedFile<E>::find_aliases(Symbol<E> *sym) {
 template <typename E>
 i64 SharedFile<E>::get_alignment(Symbol<E> *sym) {
   ElfShdr<E> &shdr = this->elf_sections[sym->esym().st_shndx];
-  i64 p2align = std::min(std::countr_zero<u64>(sym->value),
-                         std::countr_zero<u64>(shdr.sh_addralign));
-
-  // We do not want a ridiculously large alignment. Cap it arbitrary at 64.
-  return std::min(64, 1 << p2align);
+  i64 align = std::max<i64>(1, shdr.sh_addralign);
+  if (sym->value)
+    align = std::min<i64>(align, 1LL << std::countr_zero(sym->value));
+  return align;
 }
 
 template <typename E>
@@ -1514,6 +1567,8 @@ void SharedFile<E>::compute_symtab_size(Context<E> &ctx) {
   if (ctx.arg.strip_all)
     return;
 
+  this->output_sym_indices.resize(this->elf_syms.size(), -1);
+
   // Compute the size of global symbols.
   for (i64 i = this->first_global; i < this->symbols.size(); i++) {
     Symbol<E> &sym = *this->symbols[i];
@@ -1521,7 +1576,7 @@ void SharedFile<E>::compute_symtab_size(Context<E> &ctx) {
     if (sym.file == this && (sym.is_imported || sym.is_exported) &&
         (!ctx.arg.retain_symbols_file || sym.write_to_symtab)) {
       this->strtab_size += sym.name().size() + 1;
-      this->num_global_symtab++;
+      this->output_sym_indices[i] = this->num_global_symtab++;
       sym.write_to_symtab = true;
     }
   }
@@ -1532,24 +1587,21 @@ void SharedFile<E>::populate_symtab(Context<E> &ctx) {
   ElfSym<E> *symtab =
     (ElfSym<E> *)(ctx.buf + ctx.symtab->shdr.sh_offset) + this->global_symtab_idx;
 
-  u8 *strtab = ctx.buf + ctx.strtab->shdr.sh_offset + this->strtab_offset;
+  u8 *strtab = ctx.buf + ctx.strtab->shdr.sh_offset;
+  i64 strtab_off = this->strtab_offset;
 
-  for (i64 i = this->first_global; i < this->elf_syms.size(); i++) {
-    Symbol<E> &sym = *this->symbols[i];
+  for (i64 i = 0, j = this->first_global; j < this->elf_syms.size(); i++, j++) {
+    Symbol<E> &sym = *this->symbols[j];
     if (sym.file != this || !sym.write_to_symtab)
       continue;
 
-    ElfSym<E> &esym = *symtab++;
-    esym.st_name = strtab - (ctx.buf + ctx.strtab->shdr.sh_offset);
-    esym.st_value = 0;
-    esym.st_size = 0;
-    esym.st_type = STT_NOTYPE;
-    esym.st_bind = STB_GLOBAL;
-    esym.st_visibility = sym.visibility;
-    esym.st_shndx = SHN_UNDEF;
+    U32<E> *xindex = nullptr;
+    if (ctx.symtab_shndx)
+      xindex = (U32<E> *)(ctx.buf + ctx.symtab_shndx->shdr.sh_offset +
+                          (this->global_symtab_idx + i) * 4);
 
-    write_string(strtab, sym.name());
-    strtab += sym.name().size() + 1;
+    *symtab++ = to_output_esym(ctx, sym, strtab_off, xindex);
+    strtab_off += write_string(strtab + strtab_off, sym.name());
   }
 }
 
