@@ -2,9 +2,14 @@
 
 #include <limits>
 #include <zlib.h>
+#include <zstd.h>
 
 namespace mold::elf {
 
+typedef enum {
+  NONE, ERROR, COPYREL, DYN_COPYREL, PLT, CPLT, DYN_CPLT, DYNREL, BASEREL, IFUNC,
+} Action;
+
 template <typename E>
 bool CieRecord<E>::equals(const CieRecord<E> &other) const {
   if (get_contents() != other.get_contents())
@@ -19,13 +24,13 @@ bool CieRecord<E>::equals(const CieRecord<E> &other) const {
     if (x[i].r_offset - input_offset != y[i].r_offset - other.input_offset ||
         x[i].r_type != y[i].r_type ||
         file.symbols[x[i].r_sym] != other.file.symbols[y[i].r_sym] ||
-        input_section.get_addend(x[i]) != other.input_section.get_addend(y[i]))
+        get_addend(input_section, x[i]) != get_addend(other.input_section, y[i]))
       return false;
   }
   return true;
 }
 
-static inline i64 to_p2align(u64 alignment) {
+static i64 to_p2align(u64 alignment) {
   if (alignment == 0)
     return 0;
   return std::countr_zero(alignment);
@@ -38,19 +43,13 @@ InputSection<E>::InputSection(Context<E> &ctx, ObjectFile<E> &file,
   if (shndx < file.elf_sections.size())
     contents = {(char *)file.mf->data + shdr().sh_offset, (size_t)shdr().sh_size};
 
-  if (name.starts_with(".zdebug")) {
-    sh_size = *(ub64 *)&contents[4];
-    p2align = to_p2align(shdr().sh_addralign);
-    compressed = true;
-  } else if (shdr().sh_flags & SHF_COMPRESSED) {
+  if (shdr().sh_flags & SHF_COMPRESSED) {
     ElfChdr<E> &chdr = *(ElfChdr<E> *)&contents[0];
     sh_size = chdr.ch_size;
     p2align = to_p2align(chdr.ch_addralign);
-    compressed = true;
   } else {
     sh_size = shdr().sh_size;
     p2align = to_p2align(shdr().sh_addralign);
-    compressed = false;
   }
 
   // Sections may have been compressed. We usually uncompress them
@@ -60,124 +59,54 @@ InputSection<E>::InputSection(Context<E> &ctx, ObjectFile<E> &file,
   // addends are in relocations.
   if constexpr (!is_rela<E>)
     uncompress(ctx);
-
-  output_section =
-    OutputSection<E>::get_instance(ctx, name, shdr().sh_type, shdr().sh_flags);
 }
 
 template <typename E>
 void InputSection<E>::uncompress(Context<E> &ctx) {
-  if (!compressed || uncompressed)
+  if (!(shdr().sh_flags & SHF_COMPRESSED) || uncompressed)
     return;
 
   u8 *buf = new u8[sh_size];
   uncompress_to(ctx, buf);
-  contents = {(char *)buf, sh_size};
+  contents = std::string_view((char *)buf, sh_size);
   ctx.string_pool.emplace_back(buf);
   uncompressed = true;
 }
 
 template <typename E>
 void InputSection<E>::uncompress_to(Context<E> &ctx, u8 *buf) {
-  if (!compressed || uncompressed) {
+  if (!(shdr().sh_flags & SHF_COMPRESSED) || uncompressed) {
     memcpy(buf, contents.data(), contents.size());
     return;
   }
 
-  auto do_uncompress = [&](std::string_view data) {
-    unsigned long size = sh_size;
-    if (::uncompress(buf, &size, (u8 *)data.data(), data.size()) != Z_OK)
-      Fatal(ctx) << *this << ": uncompress failed";
-    assert(size == sh_size);
-  };
-
-  if (name().starts_with(".zdebug")) {
-    // Old-style compressed section
-    if (!contents.starts_with("ZLIB") || contents.size() <= 12)
-      Fatal(ctx) << *this << ": corrupted compressed section";
-    do_uncompress(contents.substr(12));
-    return;
-  }
-
-  assert(shdr().sh_flags & SHF_COMPRESSED);
-
-  // New-style compressed section
   if (contents.size() < sizeof(ElfChdr<E>))
     Fatal(ctx) << *this << ": corrupted compressed section";
 
   ElfChdr<E> &hdr = *(ElfChdr<E> *)&contents[0];
-  if (hdr.ch_type != ELFCOMPRESS_ZLIB)
+  std::string_view data = contents.substr(sizeof(ElfChdr<E>));
+
+  switch (hdr.ch_type) {
+  case ELFCOMPRESS_ZLIB: {
+    unsigned long size = sh_size;
+    if (::uncompress(buf, &size, (u8 *)data.data(), data.size()) != Z_OK)
+      Fatal(ctx) << *this << ": uncompress failed";
+    assert(size == sh_size);
+    break;
+  }
+  case ELFCOMPRESS_ZSTD:
+    if (ZSTD_decompress(buf, sh_size, (u8 *)data.data(), data.size()) != sh_size)
+      Fatal(ctx) << *this << ": ZSTD_decompress failed";
+    break;
+  default:
     Fatal(ctx) << *this << ": unsupported compression type: 0x"
                << std::hex << hdr.ch_type;
-  do_uncompress(contents.substr(sizeof(ElfChdr<E>)));
-}
-
-typedef enum { NONE, ERROR, COPYREL, PLT, CPLT, DYNREL, BASEREL } Action;
-
-template <typename E>
-static void
-dispatch(Context<E> &ctx, InputSection<E> &isec, Action action,
-         Symbol<E> &sym, const ElfRel<E> &rel) {
-  auto error = [&] {
-    std::string msg = sym.is_absolute() ? "-fno-PIC" : "-fPIC";
-    Error(ctx) << isec << ": " << rel << " relocation at offset 0x"
-               << std::hex << rel.r_offset << " against symbol `"
-               << sym << "' can not be used; recompile with " << msg;
-  };
-
-  auto check_textrel = [&] {
-    if (isec.shdr().sh_flags & SHF_WRITE)
-      return;
-
-    if (ctx.arg.z_text) {
-      error();
-    } else if (ctx.arg.warn_textrel) {
-      Warn(ctx) << isec << ": relocation against symbol `" << sym
-                << "' in read-only section";
-    }
-    ctx.has_textrel = true;
-  };
-
-  switch (action) {
-  case NONE:
-    return;
-  case ERROR:
-    error();
-    return;
-  case COPYREL:
-    if (!ctx.arg.z_copyreloc) {
-      error();
-    } else if (sym.esym().st_visibility == STV_PROTECTED) {
-      Error(ctx) << isec << ": cannot make copy relocation for protected symbol '"
-                 << sym << "', defined in " << *sym.file
-                 << "; recompile with -fPIC";
-    }
-    sym.flags |= NEEDS_COPYREL;
-    return;
-  case PLT:
-    sym.flags |= NEEDS_PLT;
-    return;
-  case CPLT:
-    sym.flags |= NEEDS_CPLT;
-    return;
-  case DYNREL:
-    assert(sym.is_imported);
-    check_textrel();
-    isec.file.num_dynrel++;
-    return;
-  case BASEREL:
-    check_textrel();
-    if (!isec.is_relr_reloc(ctx, rel))
-      isec.file.num_dynrel++;
-    return;
-  default:
-    unreachable();
   }
 }
 
 template <typename E>
-static Action get_rel_action(Context<E> &ctx, const Action table[3][4],
-                             Symbol<E> &sym) {
+static Action get_rel_action(Context<E> &ctx, Symbol<E> &sym,
+                             const Action table[3][4]) {
   auto get_output_type = [&] {
     if (ctx.arg.shared)
       return 0;
@@ -200,84 +129,253 @@ static Action get_rel_action(Context<E> &ctx, const Action table[3][4],
 }
 
 template <typename E>
-void InputSection<E>::scan_abs_rel(Context<E> &ctx, Symbol<E> &sym,
-                                   const ElfRel<E> &rel) {
-  // This is a decision table for absolute relocations that is smaller
-  // than the word size (e.g. R_X86_64_32). Since the dynamic linker
-  // generally does not support dynamic relocations smaller than the
-  // word size, we need to report an error if a relocation cannot be
-  // resolved at link-time.
-  constexpr Action table[][4] = {
-    // Absolute  Local    Imported data  Imported code
-    {  NONE,     ERROR,   ERROR,         ERROR },  // Shared object
-    {  NONE,     ERROR,   ERROR,         ERROR },  // Position-independent exec
-    {  NONE,     NONE,    COPYREL,       CPLT  },  // Position-dependent exec
+static void scan_rel(Context<E> &ctx, InputSection<E> &isec, Symbol<E> &sym,
+                     const ElfRel<E> &rel, Action action) {
+  bool writable = (isec.shdr().sh_flags & SHF_WRITE);
+
+  auto error = [&] {
+    std::string msg = sym.is_absolute() ? "-fno-PIC" : "-fPIC";
+    Error(ctx) << isec << ": " << rel << " relocation at offset 0x"
+               << std::hex << rel.r_offset << " against symbol `"
+               << sym << "' can not be used; recompile with " << msg;
   };
-  Action action = get_rel_action(ctx, table, sym);
-  dispatch(ctx, *this, action, sym, rel);
-}
 
-template <typename E>
-static Action get_abs_dyn_action(Context<E> &ctx, Symbol<E> &sym) {
-  // This is a decision table for absolute relocations for the word
-  // size data (e.g. R_X86_64_64). Unlike the above, we can emit a
-  // dynamic relocation if we cannot resolve its address at link-time.
-  constexpr Action table[][4] = {
-    // Absolute  Local    Imported data  Imported code
-    {  NONE,     BASEREL, DYNREL,        DYNREL },  // Shared object
-    {  NONE,     BASEREL, DYNREL,        DYNREL },  // Position-independent exec
-    {  NONE,     NONE,    COPYREL,       CPLT   },  // Position-dependent exec
+  auto check_textrel = [&] {
+    if (!writable) {
+      if (ctx.arg.z_text) {
+        error();
+      } else if (ctx.arg.warn_textrel) {
+        Warn(ctx) << isec << ": relocation against symbol `" << sym
+                  << "' in read-only section";
+      }
+      ctx.has_textrel = true;
+    }
   };
-  return get_rel_action(ctx, table, sym);
+
+  auto copyrel = [&] {
+    assert(sym.is_imported);
+    if (sym.esym().st_visibility == STV_PROTECTED) {
+      Error(ctx) << isec
+                 << ": cannot make copy relocation for protected symbol '" << sym
+                 << "', defined in " << *sym.file << "; recompile with -fPIC";
+    }
+    sym.flags |= NEEDS_COPYREL;
+  };
+
+  auto dynrel = [&] {
+    check_textrel();
+    isec.file.num_dynrel++;
+  };
+
+  switch (action) {
+  case NONE:
+    break;
+  case ERROR:
+    error();
+    break;
+  case COPYREL:
+    if (!ctx.arg.z_copyreloc)
+      error();
+    copyrel();
+    break;
+  case DYN_COPYREL:
+    if (writable || !ctx.arg.z_copyreloc)
+      dynrel();
+    else
+      copyrel();
+    break;
+  case PLT:
+    sym.flags |= NEEDS_PLT;
+    break;
+  case CPLT:
+    sym.flags |= NEEDS_CPLT;
+    break;
+  case DYN_CPLT:
+    if (writable)
+      dynrel();
+    else
+      sym.flags |= NEEDS_CPLT;
+    break;
+  case DYNREL:
+  case IFUNC:
+    dynrel();
+    break;
+  case BASEREL:
+    check_textrel();
+    if (!isec.is_relr_reloc(ctx, rel))
+      isec.file.num_dynrel++;
+    break;
+  default:
+    unreachable();
+  }
 }
 
 template <typename E>
-void InputSection<E>::scan_abs_dyn_rel(Context<E> &ctx, Symbol<E> &sym,
-                                       const ElfRel<E> &rel) {
-  Action action = get_abs_dyn_action(ctx, sym);
-  dispatch(ctx, *this, action, sym, rel);
-}
-
-template <typename E>
-void InputSection<E>::scan_pcrel_rel(Context<E> &ctx, Symbol<E> &sym,
-                                     const ElfRel<E> &rel) {
+static Action get_pcrel_action(Context<E> &ctx, Symbol<E> &sym) {
   // This is for PC-relative relocations (e.g. R_X86_64_PC32).
   // We cannot promote them to dynamic relocations because the dynamic
   // linker generally does not support PC-relative relocations.
-  constexpr Action table[][4] = {
+  constexpr static Action table[3][4] = {
     // Absolute  Local    Imported data  Imported code
     {  ERROR,    NONE,    ERROR,         PLT    },  // Shared object
     {  ERROR,    NONE,    COPYREL,       PLT    },  // Position-independent exec
     {  NONE,     NONE,    COPYREL,       CPLT   },  // Position-dependent exec
   };
-  Action action = get_rel_action(ctx, table, sym);
-  dispatch(ctx, *this, action, sym, rel);
+
+  return get_rel_action(ctx, sym, table);
 }
 
 template <typename E>
-void InputSection<E>::apply_abs_dyn_rel(Context<E> &ctx, Symbol<E> &sym,
-                                        const ElfRel<E> &rel, u8 *loc,
-                                        u64 S, i64 A, u64 P, ElfRel<E> *&dynrel) {
-  switch (get_abs_dyn_action(ctx, sym)) {
+static Action get_absrel_action(Context<E> &ctx, Symbol<E> &sym) {
+  // This is a decision table for absolute relocations that is smaller
+  // than the word size (e.g. R_X86_64_32). Since the dynamic linker
+  // generally does not support dynamic relocations smaller than the
+  // word size, we need to report an error if a relocation cannot be
+  // resolved at link-time.
+  constexpr static Action table[3][4] = {
+    // Absolute  Local    Imported data  Imported code
+    {  NONE,     ERROR,   ERROR,         ERROR },  // Shared object
+    {  NONE,     ERROR,   ERROR,         ERROR },  // Position-independent exec
+    {  NONE,     NONE,    COPYREL,       CPLT  },  // Position-dependent exec
+  };
+
+  return get_rel_action(ctx, sym, table);
+}
+
+template <typename E>
+static Action get_dyn_absrel_action(Context<E> &ctx, Symbol<E> &sym) {
+  if (sym.is_ifunc())
+    return IFUNC;
+
+  // This is a decision table for absolute relocations for the word
+  // size data (e.g. R_X86_64_64). Unlike the absrel_table, we can emit
+  // a dynamic relocation if we cannot resolve an address at link-time.
+  constexpr static Action table[3][4] = {
+    // Absolute  Local    Imported data  Imported code
+    {  NONE,     BASEREL, DYNREL,        DYNREL   },  // Shared object
+    {  NONE,     BASEREL, DYNREL,        DYNREL   },  // Position-independent exec
+    {  NONE,     NONE,    DYN_COPYREL,   DYN_CPLT },  // Position-dependent exec
+  };
+
+  return get_rel_action(ctx, sym, table);
+}
+
+template <typename E>
+static Action get_ppc64_toc_action(Context<E> &ctx, Symbol<E> &sym) {
+  if (sym.is_ifunc())
+    return IFUNC;
+
+  // As a special case, we do not create copy relocations nor canonical
+  // PLTs for .toc sections. PPC64's .toc is a compiler-generated
+  // GOT-like section, and no user-generated code directly uses values
+  // in it.
+  constexpr static Action table[3][4] = {
+    // Absolute  Local    Imported data  Imported code
+    {  NONE,     BASEREL, DYNREL,        DYNREL },  // Shared object
+    {  NONE,     BASEREL, DYNREL,        DYNREL },  // Position-independent exec
+    {  NONE,     NONE,    DYNREL,        DYNREL },  // Position-dependent exec
+  };
+
+  return get_rel_action(ctx, sym, table);
+}
+
+template <typename E>
+void InputSection<E>::scan_pcrel(Context<E> &ctx, Symbol<E> &sym,
+                                 const ElfRel<E> &rel) {
+  scan_rel(ctx, *this, sym, rel, get_pcrel_action(ctx, sym));
+}
+
+template <typename E>
+void InputSection<E>::scan_absrel(Context<E> &ctx, Symbol<E> &sym,
+                                  const ElfRel<E> &rel) {
+  scan_rel(ctx, *this, sym, rel, get_absrel_action(ctx, sym));
+}
+
+template <typename E>
+void InputSection<E>::scan_dyn_absrel(Context<E> &ctx, Symbol<E> &sym,
+                                      const ElfRel<E> &rel) {
+  scan_rel(ctx, *this, sym, rel, get_dyn_absrel_action(ctx, sym));
+}
+
+template <typename E>
+void InputSection<E>::scan_toc_rel(Context<E> &ctx, Symbol<E> &sym,
+                                   const ElfRel<E> &rel) {
+  scan_rel(ctx, *this, sym, rel, get_ppc64_toc_action(ctx, sym));
+}
+
+template <typename E>
+static void apply_absrel(Context<E> &ctx, InputSection<E> &isec,
+                         Symbol<E> &sym, const ElfRel<E> &rel, u8 *loc,
+                         u64 S, i64 A, u64 P, ElfRel<E> *&dynrel,
+                         Action action) {
+  bool writable = (isec.shdr().sh_flags & SHF_WRITE);
+
+  auto apply_dynrel = [&] {
+    *dynrel++ = ElfRel<E>(P, E::R_ABS, sym.get_dynsym_idx(ctx), A);
+    if (ctx.arg.apply_dynamic_relocs)
+      *(Word<E> *)loc = A;
+  };
+
+  switch (action) {
   case COPYREL:
   case CPLT:
   case NONE:
     *(Word<E> *)loc = S + A;
     break;
   case BASEREL:
-    if (!is_relr_reloc(ctx, rel))
+    if (isec.is_relr_reloc(ctx, rel)) {
+      *(Word<E> *)loc = S + A;
+    } else {
       *dynrel++ = ElfRel<E>(P, E::R_RELATIVE, 0, S + A);
-    *(Word<E> *)loc = S + A;
+      if (ctx.arg.apply_dynamic_relocs)
+        *(Word<E> *)loc = S + A;
+    }
+    break;
+  case DYN_COPYREL:
+    if (writable || !ctx.arg.z_copyreloc)
+      apply_dynrel();
+    else
+      *(Word<E> *)loc = S + A;
+    break;
+  case DYN_CPLT:
+    if (writable)
+      apply_dynrel();
+    else
+      *(Word<E> *)loc = S + A;
     break;
   case DYNREL:
-    *dynrel++ = ElfRel<E>(P, E::R_ABS, sym.get_dynsym_idx(ctx), A);
-    *(Word<E> *)loc = A;
+    apply_dynrel();
     break;
+  case IFUNC: {
+    u64 addr = sym.get_addr(ctx, NO_PLT) + A;
+    *dynrel++ = ElfRel<E>(P, E::R_IRELATIVE, 0, addr);
+    if (ctx.arg.apply_dynamic_relocs)
+      *(Word<E> *)loc = addr;
+    break;
+  }
   default:
     unreachable();
   }
 }
 
+template <typename E>
+void InputSection<E>::apply_dyn_absrel(Context<E> &ctx, Symbol<E> &sym,
+                                       const ElfRel<E> &rel, u8 *loc,
+                                       u64 S, i64 A, u64 P,
+                                       ElfRel<E> *&dynrel) {
+  apply_absrel(ctx, *this, sym, rel, loc, S, A, P, dynrel,
+               get_dyn_absrel_action(ctx, sym));
+}
+
+template <typename E>
+void InputSection<E>::apply_toc_rel(Context<E> &ctx, Symbol<E> &sym,
+                                    const ElfRel<E> &rel, u8 *loc,
+                                    u64 S, i64 A, u64 P,
+                                    ElfRel<E> *&dynrel) {
+  apply_absrel(ctx, *this, sym, rel, loc, S, A, P, dynrel,
+               get_ppc64_toc_action(ctx, sym));
+}
+
 template <typename E>
 void InputSection<E>::write_to(Context<E> &ctx, u8 *buf) {
   if (shdr().sh_type == SHT_NOBITS || sh_size == 0)
@@ -286,17 +384,17 @@ void InputSection<E>::write_to(Context<E> &ctx, u8 *buf) {
   // Copy data
   if constexpr (is_riscv<E>) {
     copy_contents_riscv(ctx, buf);
-  } else if (compressed) {
-    uncompress_to(ctx, buf);
   } else {
-    memcpy(buf, contents.data(), contents.size());
+    uncompress_to(ctx, buf);
   }
 
   // Apply relocations
-  if (shdr().sh_flags & SHF_ALLOC)
-    apply_reloc_alloc(ctx, buf);
-  else
-    apply_reloc_nonalloc(ctx, buf);
+  if (!ctx.arg.relocatable) {
+    if (shdr().sh_flags & SHF_ALLOC)
+      apply_reloc_alloc(ctx, buf);
+    else
+      apply_reloc_nonalloc(ctx, buf);
+  }
 }
 
 // Get the name of a function containin a given offset.
@@ -365,12 +463,10 @@ void report_undef_errors(Context<E> &ctx) {
   ctx.checkpoint();
 }
 
-#define INSTANTIATE(E)                                                  \
-  template struct CieRecord<E>;                                         \
-  template class InputSection<E>;                                       \
-  template void report_undef_errors(Context<E> &)
+using E = MOLD_TARGET;
 
-
-INSTANTIATE_ALL;
+template struct CieRecord<E>;
+template class InputSection<E>;
+template void report_undef_errors(Context<E> &);
 
 } // namespace mold::elf
