@@ -251,6 +251,13 @@ MachineType get_script_output_type(Context<E> &ctx, MappedFile<Context<E>> *mf)
     if (tok[2] == "elf32-i386")
       return MachineType::I386;
   }
+
+  if (tok.size() >= 3 && (tok[0] == "INPUT" || tok[0] == "GROUP") &&
+      tok[1] == "(")
+    if (MappedFile<Context<E>> *mf =
+        MappedFile<Context<E>>::open(ctx, std::string(unquote(tok[2]))))
+      return get_machine_type(ctx, mf);
+
   return MachineType::NONE;
 }
 
@@ -271,8 +278,7 @@ static bool read_label(std::span<std::string_view> &tok,
 template <typename E>
 static void
 read_version_script_commands(Context<E> &ctx, std::span<std::string_view> &tok,
-                             u16 ver_idx, bool is_cpp) {
-  ctx.version_specified = true;
+                             std::string_view ver_str, u16 ver_idx, bool is_cpp) {
   bool is_global = true;
 
   while (!tok.empty() && tok[0] != "}") {
@@ -292,11 +298,11 @@ read_version_script_commands(Context<E> &ctx, std::span<std::string_view> &tok,
       if (!tok.empty() && tok[0] == "\"C\"") {
         tok = tok.subspan(1);
         tok = skip(ctx, tok, "{");
-        read_version_script_commands( ctx, tok, ver_idx, false);
+        read_version_script_commands( ctx, tok, ver_str, ver_idx, false);
       } else {
         tok = skip(ctx, tok, "\"C++\"");
         tok = skip(ctx, tok, "{");
-        read_version_script_commands(ctx, tok, ver_idx, true);
+        read_version_script_commands(ctx, tok, ver_str, ver_idx, true);
       }
 
       tok = skip(ctx, tok, "}");
@@ -306,10 +312,13 @@ read_version_script_commands(Context<E> &ctx, std::span<std::string_view> &tok,
 
     if (tok[0] == "*") {
       ctx.default_version = (is_global ? ver_idx : VER_NDX_LOCAL);
+      ctx.default_version_from_version_script = true;
     } else if (is_global) {
-      ctx.version_patterns.push_back({unquote(tok[0]), ver_idx, is_cpp});
+      ctx.version_patterns.push_back({unquote(tok[0]), current_file<E>->name,
+                                      ver_str, ver_idx, is_cpp});
     } else {
-      ctx.version_patterns.push_back({unquote(tok[0]), VER_NDX_LOCAL, is_cpp});
+      ctx.version_patterns.push_back({unquote(tok[0]), current_file<E>->name,
+                                      ver_str, VER_NDX_LOCAL, is_cpp});
     }
 
     tok = tok.subspan(1);
@@ -325,17 +334,21 @@ void read_version_script(Context<E> &ctx, std::span<std::string_view> &tok) {
   u16 next_ver = VER_NDX_LAST_RESERVED + ctx.arg.version_definitions.size() + 1;
 
   while (!tok.empty() && tok[0] != "}") {
+    std::string_view ver_str;
     u16 ver_idx;
+
     if (tok[0] == "{") {
+      ver_str = "global";
       ver_idx = VER_NDX_GLOBAL;
     } else {
+      ver_str = tok[0];
       ver_idx = next_ver++;
       ctx.arg.version_definitions.push_back(std::string(tok[0]));
       tok = tok.subspan(1);
     }
 
     tok = skip(ctx, tok, "{");
-    read_version_script_commands(ctx, tok, ver_idx, false);
+    read_version_script_commands(ctx, tok, ver_str, ver_idx, false);
     tok = skip(ctx, tok, "}");
     if (!tok.empty() && tok[0] != ";")
       tok = tok.subspan(1);
@@ -344,10 +357,9 @@ void read_version_script(Context<E> &ctx, std::span<std::string_view> &tok) {
 }
 
 template <typename E>
-void parse_version_script(Context<E> &ctx, std::string path) {
-  current_file<E> = MappedFile<Context<E>>::must_open(ctx, path);
-  std::vector<std::string_view> vec =
-    tokenize(ctx, current_file<E>->get_contents());
+void parse_version_script(Context<E> &ctx, MappedFile<Context<E>> *mf) {
+  current_file<E> = mf;
+  std::vector<std::string_view> vec = tokenize(ctx, mf->get_contents());
   std::span<std::string_view> tok = vec;
   read_version_script(ctx, tok);
   if (!tok.empty())
@@ -357,8 +369,6 @@ void parse_version_script(Context<E> &ctx, std::string path) {
 template <typename E>
 void read_dynamic_list_commands(Context<E> &ctx, std::span<std::string_view> &tok,
                                 bool is_cpp) {
-  ctx.version_specified = true;
-
   while (!tok.empty() && tok[0] != "}") {
     if (tok[0] == "extern") {
       tok = tok.subspan(1);
@@ -381,19 +391,19 @@ void read_dynamic_list_commands(Context<E> &ctx, std::span<std::string_view> &to
     if (tok[0] == "*")
       ctx.default_version = VER_NDX_GLOBAL;
     else
-      ctx.version_patterns.push_back({unquote(tok[0]), VER_NDX_GLOBAL, is_cpp});
+      ctx.version_patterns.push_back({unquote(tok[0]), current_file<E>->name,
+                                      "global", VER_NDX_GLOBAL, is_cpp});
 
     tok = skip(ctx, tok.subspan(1), ";");
   }
 }
 
 template <typename E>
-void parse_dynamic_list(Context<E> &ctx, std::string path) {
-  current_file<E> = MappedFile<Context<E>>::must_open(ctx, path);
-  std::vector<std::string_view> vec =
-    tokenize(ctx, current_file<E>->get_contents());
-
+void parse_dynamic_list(Context<E> &ctx, MappedFile<Context<E>> *mf) {
+  current_file<E> = mf;
+  std::vector<std::string_view> vec = tokenize(ctx, mf->get_contents());
   std::span<std::string_view> tok = vec;
+
   tok = skip(ctx, tok, "{");
   read_dynamic_list_commands(ctx, tok, false);
   tok = skip(ctx, tok, "}");
@@ -403,16 +413,11 @@ void parse_dynamic_list(Context<E> &ctx, std::string path) {
     SyntaxError(ctx, tok[0]) << "trailing garbage token";
 }
 
-#define INSTANTIATE(E)                                                          \
-  template                                                                      \
-  void parse_linker_script(Context<E> &, MappedFile<Context<E>> *);             \
-  template                                                                      \
-  MachineType get_script_output_type(Context<E> &, MappedFile<Context<E>> *);   \
-  template                                                                      \
-  void parse_version_script(Context<E> &, std::string);                         \
-  template                                                                      \
-  void parse_dynamic_list(Context<E> &, std::string);
+using E = MOLD_TARGET;
 
-INSTANTIATE_ALL;
+template void parse_linker_script(Context<E> &, MappedFile<Context<E>> *);
+template MachineType get_script_output_type(Context<E> &, MappedFile<Context<E>> *);
+template void parse_version_script(Context<E> &, MappedFile<Context<E>> *);
+template void parse_dynamic_list(Context<E> &, MappedFile<Context<E>> *);
 
 } // namespace mold::elf
