@@ -28,9 +28,10 @@ int main(int argc, char **argv);
 }
 
 static std::string get_mold_version() {
+  std::string name = MOLD_IS_SOLD ? "mold (sold) " : "mold ";
   if (mold_git_hash.empty())
-    return "mold " MOLD_VERSION " (compatible with GNU ld)";
-  return "mold " MOLD_VERSION " (" + mold_git_hash + "; compatible with GNU ld)";
+    return name + MOLD_VERSION + " (compatible with GNU ld)";
+  return name + MOLD_VERSION + " (" + mold_git_hash + "; compatible with GNU ld)";
 }
 
 void cleanup() {
@@ -145,8 +146,11 @@ i64 get_default_thread_count() {
 int main(int argc, char **argv) {
   mold::mold_version = mold::get_mold_version();
 
+#if MOLD_IS_SOLD
   std::string cmd = mold::filepath(argv[0]).filename().string();
   if (cmd == "ld64" || cmd == "ld64.mold")
     return mold::macho::main(argc, argv);
+#endif
+
   return mold::elf::main(argc, argv);
 }
