@@ -1,24 +1,12 @@
 #!/bin/bash
-export LC_ALL=C
-set -e
-CC="${TEST_CC:-cc}"
-CXX="${TEST_CXX:-c++}"
-GCC="${TEST_GCC:-gcc}"
-GXX="${TEST_GXX:-g++}"
-OBJDUMP="${OBJDUMP:-objdump}"
-MACHINE="${MACHINE:-$(uname -m)}"
-testname=$(basename "$0" .sh)
-echo -n "Testing $testname ... "
-t=out/test/elf/$MACHINE/$testname
-mkdir -p $t
+. $(dirname $0)/common.inc
 
 if [ $MACHINE = x86_64 ]; then
   dialect=gnu2
 elif [ $MACHINE = aarch64 ]; then
   dialect=desc
 else
-  echo skipped
-  exit
+  skip
 fi
 
 cat <<EOF | $GCC -fPIC -mtls-dialect=$dialect -c -o $t/a.o -xc -
@@ -63,5 +51,3 @@ $QEMU $t/exe | grep -q '42 5'
 $CC -B. -shared -o $t/c.so $t/a.o -Wl,-no-relax
 $CC -B. -o $t/exe $t/b.o $t/c.so -Wl,-no-relax
 $QEMU $t/exe | grep -q '42 5'
-
-echo OK
